# Copyright (c) 2021-2022, NVIDIA CORPORATION. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import cv2
import sys
import vpi
import numpy as np
from PIL import Image
from argparse import ArgumentParser

# ----------------------------
# Parse command line arguments

parser = ArgumentParser()
parser.add_argument('backend', choices=['cpu','cuda','pva','ofa','ofa-pva-vic','pva-nvenc-vic'],
                    help='Backend to be used for processing')

parser.add_argument('left',
                    help='Rectified left input image from a stereo pair')

parser.add_argument('right',
                    help='Rectified right input image from a stereo pair')

args = parser.parse_args()

# pixel value scaling factor when loading input
scale=1

if args.backend == 'cpu':
    backend = vpi.Backend.CPU
elif args.backend == 'cuda':
    backend = vpi.Backend.CUDA
elif args.backend == 'pva':
    backend = vpi.Backend.PVA
elif args.backend == 'ofa':
    backend = vpi.Backend.OFA
elif args.backend == 'ofa-pva-vic':
    backend = vpi.Backend.OFA|vpi.Backend.PVA|vpi.Backend.VIC
else:
    assert args.backend == 'pva-nvenc-vic'
    backend = vpi.Backend.PVA|vpi.Backend.NVENC|vpi.Backend.VIC

    # For PVA+NVENC+VIC mode, 16bpp input must be MSB-aligned, which
    # is equivalent to say that it is Q8.8 (fixed-point, 8 decimals).
    scale=256

# Streams for left and right independent pre-processing
streamLeft = vpi.Stream()
streamRight = vpi.Stream()

# --------------------------------------------------------------
# Load input into a vpi.Image and convert it to grayscale, 16bpp
with vpi.Backend.CUDA:
    with streamLeft:
        left = vpi.asimage(np.asarray(Image.open(args.left))).convert(vpi.Format.Y16_ER, scale=scale)
    with streamRight:
        right = vpi.asimage(np.asarray(Image.open(args.right))).convert(vpi.Format.Y16_ER, scale=scale)

# --------------------------------------------------------------
# Preprocess input

# Block linear format is needed for pva-nvenc-vic pipeline and ofa backend
# Currently we can only convert to block-linear using VIC backend.
# The input also must be 1080p
if args.backend == 'pva-nvenc-vic':
    with vpi.Backend.VIC:
        with streamLeft:
            left = left.convert(vpi.Format.Y16_ER_BL).rescale((1920,1080))
        with streamRight:
            right = right.convert(vpi.Format.Y16_ER_BL).rescale((1920,1080))
    maxDisparity = 256
elif 'ofa' in args.backend:
    with vpi.Backend.VIC:
        with streamLeft:
            left = left.convert(vpi.Format.Y16_ER_BL).rescale((1920,1080))
        with streamRight:
            right = right.convert(vpi.Format.Y16_ER_BL).rescale((1920,1080))
    maxDisparity = 128
else:
    maxDisparity = 64

downscale = 1

if args.backend == 'cuda' or args.backend == 'ofa-pva-vic':
    # only OFA-PVA-VIC and CUDA have confidence map
    confidenceMap = vpi.Image(left.size, vpi.Format.U16)
elif args.backend == 'pva-nvenc-vic':
    # PVA-NVENC-VIC only supports 1/4 of the input size
    downscale = 4
    confidenceMap = vpi.Image((left.size[0] // downscale, left.size[1] // downscale), vpi.Format.U16)
else:
    confidenceMap = None

# Use stream left to consolidate actual stereo processing
streamStereo = streamLeft

# ---------------------------------------------
# Estimate stereo disparity
with streamStereo, backend:
    disparity = vpi.stereodisp(left, right, downscale = downscale, out_confmap=confidenceMap, window=5, maxdisp=maxDisparity)

# ---------------------------------------------
# Postprocess results and save them to disk
with streamStereo, vpi.Backend.CUDA:
    # Since OFA outputs disparities in block-linear format, we must convert them to
    # pitch-linear for consistency with the other backends.
    if args.backend == 'ofa':
        disparity = disparity.convert(vpi.Format.S16, backend = vpi.Backend.VIC)

    # Scale disparity and confidence map so that values like between 0 and 255.

    # Disparities are in Q10.5 format, so to map it to float, it gets
    # divided by 32. Then the resulting disparity range, from 0 to
    # stereo.maxDisparity gets mapped to 0-255 for proper output.
    disparity = disparity.convert(vpi.Format.U8, scale=255.0/(32*maxDisparity))

    # Apply JET colormap to turn the disparities into color, reddish hues
    # represent objects closer to the camera, blueish are farther away.
    disparityColor = cv2.applyColorMap(disparity.cpu(), cv2.COLORMAP_JET)

    # Converts to RGB for output with PIL
    disparityColor = cv2.cvtColor(disparityColor, cv2.COLOR_BGR2RGB)

    if confidenceMap:
        confidenceMap = confidenceMap.convert(vpi.Format.U8, scale=255.0/65535)

        # When pixel confidence is 0, its color in the disparity
        # output is black.
        mask = cv2.threshold(confidenceMap.cpu(), 1, 255, cv2.THRESH_BINARY)[1]
        mask = cv2.cvtColor(mask, cv2.COLOR_GRAY2BGR)
        disparityColor = cv2.bitwise_and(disparityColor, mask)

# -------------------
# Save result to disk

Image.fromarray(disparityColor).save('disparity_python'+str(sys.version_info[0])+'_'+args.backend+'.png')

if confidenceMap:
    Image.fromarray(confidenceMap.cpu()).save('confidence_python'+str(sys.version_info[0])+'_'+args.backend+'.png')
