# Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import cv2
import sys
import vpi
import numpy as np
from math import sin, cos, pi
from argparse import ArgumentParser

# ----------------------------
# Parse command line arguments

parser = ArgumentParser()
parser.add_argument('backend', choices=['cpu', 'cuda','vic'],
                    help='Backend to be used for processing')

parser.add_argument('input',
                    help='Input video to be denoised')

args = parser.parse_args();

if args.backend == 'cuda':
    backend = vpi.Backend.CUDA
elif args.backend == 'cpu':
    backend = vpi.Backend.CPU
else:
    assert args.backend == 'vic'
    backend = vpi.Backend.VIC

# -----------------------------
# Open input and output videos

inVideo = cv2.VideoCapture(args.input)

fourcc = cv2.VideoWriter_fourcc(*'MPEG')
inSize = (int(inVideo.get(cv2.CAP_PROP_FRAME_WIDTH)), int(inVideo.get(cv2.CAP_PROP_FRAME_HEIGHT)))
fps = inVideo.get(cv2.CAP_PROP_FPS)

outVideo = cv2.VideoWriter('perspwarp_python'+str(sys.version_info[0])+'_'+args.backend+'.mp4',
                            fourcc, fps, inSize)

#--------------------------------------------------------------
# Main processing loop
curFrame = 1
while True:
    print("Frame: {}".format(curFrame))
    curFrame+=1

    # Read one input frame
    ret, cvFrame = inVideo.read()
    if not ret:
        break

    # Convert it to NV12_ER format to be used by VPI
    with vpi.Backend.CUDA:
        frame = vpi.asimage(cvFrame).convert(vpi.Format.NV12_ER)

    # Calculate the transformation to be applied ------------

    # Move image's center to origin of coordinate system
    T1 = np.array([[1, 0, -frame.width/2.0],
                   [0, 1, -frame.height/2.0],
                   [0, 0, 1]])

    # Apply some time-dependent perspective transform
    v1 = sin(curFrame/30.0*2*pi/2)*0.0005
    v2 = cos(curFrame/30.0*2*pi/3)*0.0005
    P = np.array([[0.66, 0, 0],
                  [0, 0.66, 0],
                  [v1, v2, 1]])

    # Move image's center back to where it was
    T2 = np.array([[1, 0, frame.width/2.0],
                   [0, 1, frame.height/2.0],
                   [0, 0, 1]])

    # Do perspective warp using the backend passed in the command line.
    with backend:
        frame = frame.perspwarp(np.matmul(T2, np.matmul(P, T1)))

    # Convert it to RGB8 for output using the CUDA backend
    with vpi.Backend.CUDA:
        frame = frame.convert(vpi.Format.RGB8)

    # Write the denoised frame to the output video
    with frame.rlock_cpu() as data:
        outVideo.write(data)
