/*
* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*  * Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*  * Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*  * Neither the name of NVIDIA CORPORATION nor the names of its
*    contributors may be used to endorse or promote products derived
*    from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <opencv2/core/version.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <vpi/OpenCVInterop.hpp>

#include <vpi/Event.h>
#include <vpi/Image.h>
#include <vpi/Status.h>
#include <vpi/Stream.h>
#include <vpi/algo/BilateralFilter.h>
#include <vpi/algo/ConvertImageFormat.h>
#include <vpi/algo/ImageFlip.h>

#include <cstring> // for memset
#include <iostream>
#include <sstream>

#define CHECK_STATUS(STMT)                                    \
    do                                                        \
    {                                                         \
        VPIStatus status = (STMT);                            \
        if (status != VPI_SUCCESS)                            \
        {                                                     \
            char buffer[VPI_MAX_STATUS_MESSAGE_LENGTH];       \
            vpiGetLastStatusMessage(buffer, sizeof(buffer));  \
            std::ostringstream ss;                            \
            ss << vpiStatusGetName(status) << ": " << buffer; \
            throw std::runtime_error(ss.str());               \
        }                                                     \
    } while (0);

int main(int argc, char *argv[])
{
    // OpenCV image that will be wrapped by a VPIImage.
    // Define it here so that it's destroyed *after* wrapper is destroyed
    cv::Mat cvImage;

    // VPI objects that will be used
    VPIImage imageBGR    = NULL;
    VPIImage inParent    = NULL;
    VPIImage outParent   = NULL;
    VPIEvent parentEvent = NULL;
    VPIImage inViews[4]  = {};
    VPIImage outViews[4] = {};
    VPIStream streams[4] = {};

    int retval = 0;

    // Views on Tegra devices need parent images to have only CPU or CUDA backends enabled.
    int imgFlags = VPI_BACKEND_CPU | VPI_BACKEND_CUDA;

    try
    {
        if (argc != 2)
        {
            throw std::runtime_error(std::string("Usage: ") + argv[0] + " <input image>");
        }

        std::string strInputFileName = argv[1];

        // Load the input image
        cvImage = cv::imread(strInputFileName);
        if (cvImage.empty())
        {
            throw std::runtime_error("Can't open '" + strInputFileName + "'");
        }

        // Wrap the loaded image into a VPIImage object to be used by VPI
        CHECK_STATUS(vpiImageCreateWrapperOpenCVMat(cvImage, 0, &imageBGR));

        // Retrieve the width and height of the loaded parent image
        const int parentWidth = cvImage.cols, parentHeight = cvImage.rows;

        // Calculate the width and height of the child image views
        const int viewWidth = parentWidth / 2, viewHeight = parentHeight / 2;

        if ((viewWidth == 0) || (viewHeight == 0))
        {
            throw std::runtime_error(std::string("Input image too small, it must be at least 2x2"));
        }

        // Create the input parent image as a single unsigned 8-bit channel
        CHECK_STATUS(vpiImageCreate(parentWidth, parentHeight, VPI_IMAGE_FORMAT_U8, imgFlags, &inParent));

        // Create the output parent image, single unsigned 8-bit channel
        CHECK_STATUS(vpiImageCreate(parentWidth, parentHeight, VPI_IMAGE_FORMAT_U8, imgFlags, &outParent));

        // Create 1 event for the parent image
        CHECK_STATUS(vpiEventCreate(0, &parentEvent));

        for (int i = 0; i < 4; ++i)
        {
            // Create 4 streams to execute algorithms on
            CHECK_STATUS(vpiStreamCreate(0, &streams[i]));

            // Define the clip bounds to be the rectangular region of each view inside the parent image
            VPIRectangleI clipBounds;

            clipBounds.x      = static_cast<int>(i % 2) * viewWidth;
            clipBounds.y      = static_cast<int>(i / 2) * viewHeight;
            clipBounds.width  = viewWidth;
            clipBounds.height = viewHeight;

            // Create each input child image view
            CHECK_STATUS(vpiImageCreateView(inParent, &clipBounds, imgFlags, &inViews[i]));

            // Create each output child image view
            CHECK_STATUS(vpiImageCreateView(outParent, &clipBounds, imgFlags, &outViews[i]));
        }

        // Run one algorithm on each view:
        // Each algorithm runs on a different stream, potentially in parallel, all streams working together
        // consuming the same input to produce a single output, using CPU and CUDA backends

        // Convert the loaded parent image to grayscale
        CHECK_STATUS(vpiSubmitConvertImageFormat(streams[0], VPI_BACKEND_CPU, imageBGR, inParent, NULL));

        // Record the parent event announcing the conversion above has finished
        CHECK_STATUS(vpiEventRecord(parentEvent, streams[0]));

        // Wait the conversion above to finish, then release all streams for view processing
        CHECK_STATUS(vpiStreamWaitEvent(streams[1], parentEvent));
        CHECK_STATUS(vpiStreamWaitEvent(streams[2], parentEvent));
        CHECK_STATUS(vpiStreamWaitEvent(streams[3], parentEvent));

        // Submit an algorithm on 1st view
        CHECK_STATUS(
            vpiSubmitBilateralFilter(streams[0], VPI_BACKEND_CPU, inViews[0], outViews[0], 3, 45, 35, VPI_BORDER_ZERO));

        // Submit an algorithm on 2nd view
        CHECK_STATUS(vpiSubmitImageFlip(streams[1], VPI_BACKEND_CPU, inViews[1], outViews[1], VPI_FLIP_HORIZ));

        // Submit an algorithm on 3rd view
        CHECK_STATUS(vpiSubmitImageFlip(streams[2], VPI_BACKEND_CUDA, inViews[2], outViews[2], VPI_FLIP_VERT));

        // Submit an algorithm on 4th view
        CHECK_STATUS(vpiSubmitImageFlip(streams[3], VPI_BACKEND_CUDA, inViews[3], outViews[3], VPI_FLIP_BOTH));

        // Synchronize all streams to make sure all views finished processing
        CHECK_STATUS(vpiStreamSync(streams[0]));
        CHECK_STATUS(vpiStreamSync(streams[1]));
        CHECK_STATUS(vpiStreamSync(streams[2]));
        CHECK_STATUS(vpiStreamSync(streams[3]));

        // Retrieve the output parent image contents and write it to disk

        // Lock output parent image to retrieve its data, checking it consists of
        // host-accessible memory buffers in pitch-linear layout
        VPIImageData outData;
        CHECK_STATUS(vpiImageLockData(outParent, VPI_LOCK_READ, VPI_IMAGE_BUFFER_HOST_PITCH_LINEAR, &outData));
        assert(outData.bufferType == VPI_IMAGE_BUFFER_HOST_PITCH_LINEAR);

        cv::Mat cvOut;
        vpiImageDataExportOpenCVMat(outData, &cvOut);
        imwrite("output_views.png", cvOut);

        // Done handling output parent image, don't forget to unlock it
        CHECK_STATUS(vpiImageUnlock(outParent));
    }
    catch (std::exception &e)
    {
        std::cerr << e.what() << std::endl;
        retval = 1;
    }

    // Clean up

    // Make sure all streams are synchronized before destroying the objects possibly in use
    for (int i = 0; i < 4; ++i)
    {
        if (streams[i] != NULL)
        {
            vpiStreamSync(streams[i]);
        }
    }

    // Make sure to destroy all views before destroying their parents
    for (int i = 0; i < 4; ++i)
    {
        vpiStreamDestroy(streams[i]);

        vpiImageDestroy(inViews[i]);
        vpiImageDestroy(outViews[i]);
    }

    vpiEventDestroy(parentEvent);

    vpiImageDestroy(imageBGR);
    vpiImageDestroy(inParent);
    vpiImageDestroy(outParent);

    return retval;
}
