# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import vpi
import numpy as np
from PIL import Image
from argparse import ArgumentParser

# Parse command line arguments
parser = ArgumentParser()
parser.add_argument('input', help='Image to be used as input')

args = parser.parse_args();

# Load input parent image into a vpi.Image
input = vpi.asimage(np.asarray(Image.open(args.input)))

# Create 4 streams for independent processing
streams  = []
for _ in range(4):
    streams.append(vpi.Stream())

# Create input parent image and output as grayscale images in the CPU using numpy
np_parent = np.zeros((input.height, input.width), np.uint8)
np_output = np.zeros((input.height, input.width), np.uint8)

# Wrap the input parent and output using VPI, this is important
# because views on Tegra devices can only be created on wrapped CPU or
# CUDA buffers.  If an image is created with vpi.Image, and there is a
# different backend available, the view cannot be created.
parent = vpi.asimage(np_parent)
output = vpi.asimage(np_output)

# Convert input parent image to grayscale
with vpi.Backend.CPU, streams[0]:
    input.convert(out=parent)

# Calculate views size: width x height
viewSize = (input.width // 2, input.height // 2)

# Define clip bounds for each view
clipBounds = vpi.RectangleI(0, 0, viewSize[0], viewSize[1])

# Create 4 input and output child image views
parViews = []
outViews = []
for i in range(4):
    clipBounds.x = (i % 2) * viewSize[0]
    clipBounds.y = (i // 2) * viewSize[1]
    parViews.append(parent.view(clipBounds))
    outViews.append(output.view(clipBounds))

# Run one algorithm on each view:
# Each algorithm runs on a different stream, potentially in parallel, all streams working together
# consuming the same input to produce a single output, using CPU and CUDA backends

with vpi.Backend.CPU:
    parViews[0].bilateral_filter(3, 45, 35, out=outViews[0], border=vpi.Border.ZERO, stream=streams[0])
    parViews[1].image_flip(vpi.Flip.HORIZ, out=outViews[1], stream=streams[1])

with vpi.Backend.CUDA:
    parViews[2].image_flip(vpi.Flip.VERT, out=outViews[2], stream=streams[2])
    parViews[3].image_flip(vpi.Flip.BOTH, out=outViews[3], stream=streams[3])

# Sync all streams to guarantee they finish before reading the output image
for i in range(4):
    streams[i].sync()

# Save result to disk
Image.fromarray(output.cpu()).save('output_views_python'+str(sys.version_info[0])+'.png')
