# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import vpi
import torch
import numpy as np

from PIL import Image, ImageOps
from argparse import ArgumentParser

# Parse command line arguments
parser = ArgumentParser()
parser.add_argument('input', help='Image to be used as input')

args = parser.parse_args()

# Make sure CUDA is available for this example
assert torch.cuda.is_available()
cuda_device = torch.device('cuda')

# Read the input image and convert it to grayscale
pil_image = ImageOps.grayscale(Image.open(args.input))

# Pillow -> NumPy --------------------------------------
np_image = np.asarray(pil_image)

# NumPy -> PyTorch/CUDA --------------------------------
torch_image = torch.asarray(np_image).cuda()

# Perform an operation using PyTorch
torch_image = torch_image/255 * 0.5

# PyTorch/CUDA -> VPI, no copies involved --------------
vpi_image = vpi.asimage(torch_image)

# Peform operations using VPI's CUDA backend
with vpi.Backend.CUDA:
    # Blur the input image with box filter
    vpi_image = vpi_image.box_filter(3)

# VPI -> PyTorch/CUDA, no copies involved --------------
with vpi_image.rlock_cuda() as cuda_buffer:
    # Perform another operation using PyTorch
    torch_tensor = torch.as_tensor(cuda_buffer, device=cuda_device)
    torch_image = torch_tensor*255 + 128

# PyTorch -> NumPy -------------------------------------
np_image = np.asarray(torch_image.cpu())

# NumPy -> Pillow --------------------------------------
pil_image = Image.fromarray(np_image)

# Save the output to disk
pil_image.convert('L').save('vpi_pytorch.png')
